/****************************************************************************
**
** Copyright (C) 2012 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (qt-info@nokia.com)
**
** This file is part of the examples of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:BSD$
** You may use this file under the terms of the BSD license as follows:
**
** "Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are
** met:
**   * Redistributions of source code must retain the above copyright
**     notice, this list of conditions and the following disclaimer.
**   * Redistributions in binary form must reproduce the above copyright
**     notice, this list of conditions and the following disclaimer in
**     the documentation and/or other materials provided with the
**     distribution.
**   * Neither the name of Nokia Corporation and its Subsidiary(-ies) nor
**     the names of its contributors may be used to endorse or promote
**     products derived from this software without specific prior written
**     permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
** "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
** LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
** A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
** OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
** SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
** LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
** $QT_END_LICENSE$
**
****************************************************************************/

#ifndef TREEMODEL_H
#define TREEMODEL_H

#include <QAbstractItemModel>
#include <QModelIndex>
#include <QVariant>
#include "Tree.h"

#define DATA_COL 0

class TreeItem;

/**
 * @brief The TreeModelInterface class is used by the tree views. It extends the QAbstractItemModel
 * with Tree and TreeItem specific functions.
 * The Interface is necessary to allow to exchange the TreeModel with the TreeModelProxy and vice versa.
 */
class TreeModelInterface
{
public:
    /** returns the model which is used by the tree view to access the data of the Tree class */
    virtual QAbstractItemModel*
    getModel() = 0;

    virtual Tree*
    getTree() const = 0;

    virtual TreeItem*
    getTreeItem( const QModelIndex& idx ) const = 0;

    virtual QModelIndex
    find( TreeItem* searchItem ) const = 0;

    virtual QModelIndexList
    find( QRegExp regExp ) const = 0;

    /** set items which should be marked as found in the view */
    virtual void
    setFoundItems( const QModelIndexList& indexList ) = 0;

    /** unmark all items in the view */
    virtual void
    clearFoundItems() = 0;

    virtual void
    setFilter( double, bool dynamic = true )
    {
        ( void )dynamic;
    }
    virtual void
    disableFilter()
    {
    }

    /** notifies the view that all values have been changed */
    virtual void
    updateValues() = 0;
};

/**
 * @brief The TreeModel class is the model for the data in the Tree class. With this model, the data of the
 * tree can be displayed with a QTreeView.
 * The model also contains tree manipulaton methods.
 */
class TreeModel : public QAbstractItemModel, public TreeModelInterface
{
    //  Q_OBJECT
public:
    TreeModel( Tree* tree );
    ~TreeModel();

    // ----- begin TreeModelInterface implementation ----

    virtual QAbstractItemModel*
    getModel();

    virtual Tree*
    getTree() const;

    virtual TreeItem*
    getTreeItem( const QModelIndex& idx ) const;

    virtual QModelIndex
    find( TreeItem* searchItem ) const;

    virtual QModelIndexList
    find( QRegExp regExp ) const;

    virtual void
    setFoundItems( const QModelIndexList& value );

    virtual void
    clearFoundItems();

    // ----- end TreeModelInterface implementation ----

    // ----- virtual methods that are required for Qt TreeView ----
    QVariant
    data( const QModelIndex& index,
          int                role ) const;
    QModelIndex
    index( int                row,
           int                column,
           const QModelIndex& parent = QModelIndex() ) const;
    QModelIndex
    parent( const QModelIndex& index ) const;
    int
    rowCount( const QModelIndex& parent = QModelIndex() ) const;
    int
    columnCount( const QModelIndex& = QModelIndex() ) const;
    Qt::ItemFlags
    flags( const QModelIndex& index ) const;

    // ----- end methods required for Qt TreeView ----

    // ----- start tree manipulation methods
    void
    replaceSubtree( TreeItem* oldItem,
                    TreeItem* newItem );

    /** Inserts newItem into to the tree as child of parent. If parent is null, the item
        is added as top level entry */
    void
    addSubtree( TreeItem* newItem,
                TreeItem* parent = 0 );

    /** removes item and all its children from the tree */
    void
    removeSubtree( TreeItem* item );

    /** removes all children of item from the tree */
    void
    removeChildren( TreeItem* item );

    /** notifies the view that all values have been changed */
    void
    updateValues();

    // ----- end tree manipulation methods
private:
    void
    drawColorIcon( QPixmap&        pix,
                   const TreeItem* item ) const;

    Tree*           tree;
    QModelIndexList foundItems;   // items which should be marked as found in the view
    QModelIndexList foundParents; // parent items of the found items
};

#endif
